#define DRR_HUGE_DOUBLE 1e10
#define DRR_LEN_TOLERANCE 1e-6
#define DRR_STRIDE_TOLERANCE 1e-10

#define OUTPUT_FORMAT_PFM 0
#define OUTPUT_FORMAT_PGM 1
#define OUTPUT_FORMAT_RAW 2

#define MSD_NUM_BINS 60

enum point_location {
	POINTLOC_LEFT,
	POINTLOC_INSIDE,
	POINTLOC_RIGHT,
};
typedef enum point_location Point_location;

struct volume_limit {
	float limits[2];	// upper and lower limits of volume, including tolerances
	int dir;		// dir == 0 if limits go from low to high
};
typedef struct volume_limit Volume_limit;

void vec3_add2 (
	__local float *v1,
	__local const float *v2
){
	v1[0] += v2[0];
	v1[1] += v2[1];
	v1[2] += v2[2];
}

void vec3_add3 (
	__local float *v1,
	__local const float *v2,
	__local const float *v3
){
	v1[0] = v2[0] + v3[0];
	v1[1] = v2[1] + v3[1];
	v1[2] = v2[2] + v3[2];
}

void vec3_copy (
	__local float *v1,
	__local const float *v2
){
	v1[0] = v2[0];
	v1[1] = v2[1];
	v1[2] = v2[2];
}

float vec3_dot (
	__local const float *v1,
	__local const float *v2
){
	return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2];
}

void vec3_scale2 (
	__local float *v1,
	__local float a
){
	v1[0] *= a;
	v1[1] *= a;
	v1[2] *= a;
}

void vec3_scale3 (
	__local float *v1,
	__local const float *v2,
	__local float a
){
	v1[0] = a * v2[0];
	v1[1] = a * v2[1];
	v1[2] = a * v2[2];
}

void vec3_sub2 (
	__local float *v1,
	__local const float *v2
){
	v1[0] -= v2[0];
	v1[1] -= v2[1];
	v1[2] -= v2[2];
}

void vec3_sub3 (
	__local float *v1,
	__local const float *v2,
	__local const float *v3
){
	v1[0] = v2[0] - v3[0];
	v1[1] = v2[1] - v3[1];
	v1[2] = v2[2] - v3[2];
}

float vec3_len (
	__local const float *v1
){
    return sqrt(vec3_dot(v1, v1));
}

void vec3_normalize (
	__local float *v1
){
    vec3_scale2(v1, 1 / vec3_len(v1));
}

float vec3_dist (
	__local const float *v1,
	__local const float *v2
){
    __local float tmp[3];
    vec3_sub3 (tmp, v1, v2);
    return vec3_len(tmp);
}

float attenuation_lookup_hu (
	float pix_density
){
	if (pix_density <= -800.0) {
		return 0.0;
	} else {
		return pix_density * 0.000022 + 0.022;
	}
}

__kernel void preprocess_attenuation_cl (
	__global float *img,
	int nvoxels
){
	int p = get_global_id(0);

	if (p < nvoxels) {
		img[p] = attenuation_lookup_hu(img[p]);
	}
}

unsigned char bin_multispectral (
	__local short pix_density
){
	const short density_min = -800;
	const unsigned short pix_divisor = 1800 / MSD_NUM_BINS;
	
	pix_density -= density_min;
	if (pix_density < 0)
		return 0;
	pix_density /= pix_divisor;
	if (pix_density >= MSD_NUM_BINS)
		return MSD_NUM_BINS - 1;
	
	return (unsigned char) pix_density;
}

void accumulate_multispectral(
	__global float *dev_multispectral,
	long img_idx,
	float pix_len,
	short pix_density
){
	__global float *bins = &dev_multispectral[img_idx * MSD_NUM_BINS];
	bins[bin_multispectral(pix_density)] += pix_len;
}

Point_location drr_test_boundary (
	__constant Volume_limit *vol_limit,
	__local float x
){
	if (vol_limit->dir == 0) {
		if (x < vol_limit->limits[0]) {
			return POINTLOC_LEFT;
		} else if (x > vol_limit->limits[1]) {
			return POINTLOC_RIGHT;
		} else {
			return POINTLOC_INSIDE;
		}
	} else {
		if (x > vol_limit->limits[0]) {
			return POINTLOC_LEFT;
		} else if (x < vol_limit->limits[1]) {
			return POINTLOC_RIGHT;
		} else {
			return POINTLOC_INSIDE;
		}
	}
}

void drr_trace_init_loopvars_nointerp (
	__local int *ai,
	__local int *aidir,
	__local float *ao,
	__local float *al,
	float pt,
	float ry,
	float offset,
	float samp
){
	if (ry > 0) {
		*aidir = 1;
		*ai = (int)floor((pt - offset + 0.5 * samp) / samp);
		*ao = samp - ((pt - offset + 0.5 * samp) - (*ai) * samp);
	} else {
		*aidir = -1;
		*ai = (int) floor((pt - offset + 0.5 * samp) / samp);
		*ao = samp - ((*ai + 1) * samp - (pt - offset + 0.5 * samp));
	}
	if (fabs(ry) > DRR_STRIDE_TOLERANCE) {
		*ao = *ao / fabs(ry);
		*al = samp / fabs(ry);
	} else {
		*ao = DRR_HUGE_DOUBLE;
		*al = DRR_HUGE_DOUBLE;
	}
}

int drr_trace_init (
	__local int *ai_x,
	__local int *ai_y,
	__local int *ai_z,
	__local int *aixdir,
	__local int *aiydir,
	__local int *aizdir,
	__local float *ao_x,
	__local float *ao_y,
	__local float *ao_z,
	__local float *al_x,
	__local float *al_y,
	__local float *al_z,
	__local float *len,
	__constant float *offset,
	__constant float *pix_spacing,
	__constant Volume_limit *vol_limits,
	__constant float *p1,
	__local float *p2
){
	Point_location ploc[3][2];
	float alpha[3][2];
	float alpha_in, alpha_out;
	__local float ray[3];
	__local float ips[2][4];
	
	/* For each of 3 dimensions: x, y, and z */
	for (int d = 0; d < 3; d++) {
		ploc[d][0] = drr_test_boundary(&vol_limits[d], p1[d]);
		ploc[d][1] = drr_test_boundary(&vol_limits[d], p2[d]);
		/* Reject segments which don't intersect the volume in this dimension */
		if (ploc[d][0] == POINTLOC_LEFT && ploc[d][1] == POINTLOC_LEFT)
			return 0;
		if (ploc[d][0] == POINTLOC_RIGHT && ploc[d][1] == POINTLOC_RIGHT)
			return 0;
	}
	
	/* If we made it here, all three dimensions have some range of alpha
	 where they intersects the volume.  However, these alphas might 
	 not overlap.  We compute the alphas, then test overlapping 
	 alphas to find the segment range within the volume.  */
	for (int d = 0; d < 3; d++) {
		if (ploc[d][0] == POINTLOC_LEFT) {
			alpha[d][0] = (vol_limits[d].limits[0] - p1[d]) / (p2[d] - p1[d]);
		} else if (ploc[d][0] == POINTLOC_RIGHT) {
			alpha[d][0] = (p1[d] - vol_limits[d].limits[1]) / (p1[d] - p2[d]);
		} else {
			alpha[d][0] = 0.0;
		}
		if (ploc[d][1] == POINTLOC_LEFT) {
			alpha[d][1] = (vol_limits[d].limits[0] - p1[d]) / (p2[d] - p1[d]);
		} else if (ploc[d][1] == POINTLOC_RIGHT) {
			alpha[d][1] = (p1[d] - vol_limits[d].limits[1]) / (p1[d] - p2[d]);
		} else {
			alpha[d][1] = 1.0;
		}
	}
	
	/* alpha_in is the alpha where the segment enters the boundary, and 
	 alpha_out is where it exits the boundary.  */
	alpha_in = alpha[0][0];
	alpha_out = alpha[0][1];
	for (int d = 1; d < 3; d++) {
		if (alpha_in < alpha[d][0])
			alpha_in = alpha[d][0];
		if (alpha_out > alpha[d][1])
			alpha_out = alpha[d][1];
	}
	
	/* If exit is before entrance, the segment does not intersect the volume */
	if (alpha_out - alpha_in < DRR_LEN_TOLERANCE) {
		return 0;
	}
	
	/* Create the volume intersection points */
	ray[0] = p2[0] - p1[0];
	ray[1] = p2[1] - p1[1];
	ray[2] = p2[2] - p1[2];
	for (int d = 0; d < 3; d++) {
		ips[0][d] = p1[d] + alpha_in * ray[d];
		ips[1][d] = p1[d] + alpha_out * ray[d];
	}
	vec3_normalize(ray);

	/* We'll go from p1 to p2 */
	/* Variable notation:
	 ai_x	// index of x
	 aixdir  // x indices moving up or down?
	 ao_x	// absolute length to next voxel crossing
	 al_x	// length between voxel crossings
	 */
	drr_trace_init_loopvars_nointerp(ai_x, aixdir, ao_x, al_x, ips[0][0], ray[0], offset[0], pix_spacing[0]);
	drr_trace_init_loopvars_nointerp(ai_y, aiydir, ao_y, al_y, ips[0][1], ray[1], offset[1], pix_spacing[1]);
	drr_trace_init_loopvars_nointerp(ai_z, aizdir, ao_z, al_z, ips[0][2], ray[2], offset[2], pix_spacing[2]);
	
	*len = vec3_dist(&ips[0][0], &ips[1][0]);

	return 1;
}

float drr_trace_ray_nointerp_2009 (
	__global float *dev_vol,
	__global float *dev_multispectral,
	__global bool *dev_valid,
	__constant int *vol_dim,
	__constant float *offset,
	__constant float *pix_spacing,
	__constant Volume_limit *vol_limits,
	__constant float *p1in,
	__local float *p2in,
	int msd_fp,
	int preprocess_attenuation,
	long img_idx
){
	__local int ai_x, ai_y, ai_z;
	__local int aixdir, aiydir, aizdir;
	__local float ao_x, ao_y, ao_z;
	__local float al_x, al_y, al_z;
	__local float len;
	float aggr_len = 0.0;
	float accum = 0.0;
	float msd_bins[MSD_NUM_BINS];
	
	if (!drr_trace_init(&ai_x, &ai_y, &ai_z, &aixdir, &aiydir, &aizdir, &ao_x, &ao_y, &ao_z, &al_x, &al_y, &al_z, &len, offset, pix_spacing, vol_limits, p1in, p2in)) {
		return 0.0;
	}

	/* We'll go from p1 to p2 */
	/* Variable notation:
	 ai_x    // index of x
	 aixdir  // x indices moving up or down?
	 ao_x    // absolute length to next voxel crossing
	 al_x    // length between voxel crossings
	 */

	long x;
	float pix_density;
	float pix_len;

	do {
		x = ai_z * vol_dim[0] * vol_dim[1] + ai_y * vol_dim[0] + ai_x;
		
		pix_density = dev_vol[x];
		
		if ((ao_x < ao_y) && (ao_x < ao_z)) {
			pix_len = ao_x;
			aggr_len += ao_x;
			ao_y -= ao_x;
			ao_z -= ao_x;
			ao_x = al_x;
			ai_x += aixdir;
		} else if (ao_y < ao_z) {
			pix_len = ao_y;
			aggr_len += ao_y;
			ao_x -= ao_y;
			ao_z -= ao_y;
			ao_y = al_y;
			ai_y += aiydir;
		} else {
			pix_len = ao_z;
			aggr_len += ao_z;
			ao_x -= ao_z;
			ao_y -= ao_z;
			ao_z = al_z;
			ai_z += aizdir;
		}

		if (preprocess_attenuation)
			accum += pix_len * pix_density;
		else
			accum += pix_len * attenuation_lookup_hu(pix_density);

		if (msd_fp) {
			accumulate_multispectral(dev_multispectral, img_idx, pix_len, pix_density);
		}
	} while (aggr_len + DRR_LEN_TOLERANCE < len);

	dev_valid[img_idx] = true;
	return accum;
}

__kernel void kernel_drr (
	__global float *dev_vol,
	__global float *dev_img,
	__global float *dev_multispectral,
	__global bool *dev_valid,
	__constant int *vol_dim,
	__constant int2 *img_dim,
	__constant float *offset, 
	__constant float *pix_spacing,
	__constant Volume_limit *vol_limits,
	__constant float *p1,
	__constant float *ul_room,
	__constant float *incr_r,
	__constant float *incr_c,
	__constant int4 *ndevice,
	float scale,
	int output_format,
	int msd_fp,
	int preprocess_attenuation,
	int exponential_mapping,
	int pixel_offset
 ){
	uint c = get_global_id(0);
	uint r = get_global_id(1);

	if (c >= (*ndevice).x || r >= (*ndevice).y)
		return;

	// Index row major into the image
	long img_idx = c + (r * (*img_dim).x);
	r += pixel_offset;

	float r_tgt[3];
	r_tgt[0] = ul_room[0] + (incr_r[0] * r);
	r_tgt[1] = ul_room[1] + (incr_r[1] * r);
	r_tgt[2] = ul_room[2] + (incr_r[2] * r);

	__local float p2[3];
	p2[0] = r_tgt[0] + incr_c[0] * c;
	p2[1] = r_tgt[1] + incr_c[1] * c;
	p2[2] = r_tgt[2] + incr_c[2] * c;

	float value = drr_trace_ray_nointerp_2009(dev_vol, dev_multispectral, dev_valid, vol_dim, offset, pix_spacing, vol_limits, p1, p2, msd_fp, preprocess_attenuation, img_idx);

	// Translate from pixels to cm*gm
	value *= 0.1;
	
	if (exponential_mapping) {
		value = exp(-value);
	}

	// User requested scaling
	value *= scale;

	if (output_format == OUTPUT_FORMAT_PGM) {
		// PGM output
		if (exponential_mapping) {
			value *= 65536;
		} else {
			value *= 15000;
		}
		if (value > 65536) {
			value = 65536;
		} else if (value < 0) {
			value = 0;
		}
	}

	dev_img[img_idx] = value;
}
