/* 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
 
 Title : OpenCL implementation of DRR Algorithm
 
 Authors : Atapattu, Chatura
 Houck, Dustin
 O'Brien, Ronald
 Partel, Michael
 
 Description : 
 
 Created : 01/13/2010
 
 Modified : ongoing
 
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
 */

// Include project header
#include "drr_opencl.h"

// Globals
cl_int error;					// Use for error checking
cl_context context;				// Context from device
cl_command_queue command_queue[MAX_GPU_COUNT];	// Command Queue from Context
cl_program program;				// Program from .cl file
cl_uint device_count;			// number of devices available
cl_device_id *devices;			// pointer to devices
cl_device_id device;			// object for individual device in for loop
char device_name[MAX_GPU_COUNT][256];	// device names


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void accumulate_multispectral (float* bins, double pix_len, short pix_density)
{
	bins[bin_multispectral(pix_density)] += pix_len;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
float attenuation_lookup_hu (float pix_density)
{
	const double min_hu = -800.0;
	const double mu_h2o = 0.022;
	if (pix_density <= min_hu) {
		return 0.0;
	} else {
		return (pix_density/1000.0) * mu_h2o + mu_h2o;
	}
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
float attenuation_lookup (float pix_density)
{
	return attenuation_lookup_hu (pix_density);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
unsigned char bin_multispectral (short pix_density)
{
	const short density_min = -800;
	const unsigned short pix_divisor = 1800 / MSD_NUM_BINS;
	
	pix_density -= density_min;
	if (pix_density < 0) return 0;
	pix_density /= pix_divisor;
	if (pix_density >= MSD_NUM_BINS) return MSD_NUM_BINS-1;
	return (unsigned char) pix_density;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void drr_render_volume_perspective (Volume* vol, double* cam, double* tgt, double* vup, double sid, double* ic, double* ps, int* ires, char* image_fn, char* multispectral_fn,  MGHDRR_Options* options)
{
	int d;
	double nrm[3];
	double vrt[3];
	double p1[3], p2[3];
	//	int res_r = ires[0];
	//	int res_c = ires[1];
	int res_r = options->image_window[1] - options->image_window[0] + 1;
	int res_c = options->image_window[3] - options->image_window[2] + 1;
	double pp_r = - ps[0];
	double pp_c = ps[1];
	double ic_room[3];
	double ul_room[3];
	double incr_r[3];
	double incr_c[3];
	double r_tgt[3];
	double tmp[3];
	Volume_limit vol_limits[3];
	
	FILE *pgm_fp, *msd_fp;
	int r, c;
	double value;
	
	/* Compute image coordinate sys (nrm,vup,vrt) relative to room coords.
	---------------
	nrm = tgt - cam
	vrt = nrm x vup
	vup = vrt x nrm
	---------------
	*/
	vec3_sub3 (nrm, tgt, cam);
	vec3_normalize1 (nrm);
	vec3_cross (vrt, nrm, vup);
	vec3_normalize1 (vrt);
	vec3_cross (vup, vrt, nrm);
	vec3_normalize1 (vup);
	
	/* Compute position of image center in room coordinates */
	vec3_scale3 (tmp, nrm, sid);
	vec3_add3 (ic_room, cam, tmp);
#if defined (VERBOSE)
	printf ("icr: ");
	vec3_print_eol (stdout, ic_room);
#endif
	
	/* Compute base and increments for pixels. Base is at upper left, 
	and increment with increasing rows and columns. */
	vec3_scale3 (incr_r, vup, pp_r);
	vec3_scale3 (incr_c, vrt, pp_c);
	vec3_copy (ul_room, ic_room);
	vec3_scale3 (tmp, incr_r, -ic[0]);
	vec3_add2 (ul_room, tmp);
	vec3_scale3 (tmp, incr_c, -ic[1]);
	vec3_add2 (ul_room, tmp);
	
	/* drr_trace_ray uses p1 & p2, p1 is the camera, p2 is in the 
	direction of the ray */
	vec3_copy (p1, cam);
	
	pgm_fp = fopen(image_fn,"wb");
	if (!pgm_fp) {
		fprintf (stderr, "Error opening %s for write\n", image_fn);
		exit (-1);
	}
	if (options->multispectral) {
		msd_fp = fopen(multispectral_fn,"wb");
		if (!msd_fp) {
			fprintf (stderr, "Error opening %s for write\n", multispectral_fn);
			exit (-1);
		}
	} else {
		msd_fp = 0;
	}
	
	if (options->output_format == OUTPUT_FORMAT_PFM) {
		fprintf (pgm_fp, 
				"Pf\n"
				"%d %d\n"
				"-1\n",
				res_c, res_r);
	} 
	else if (options->output_format == OUTPUT_FORMAT_PGM) {
		fprintf (pgm_fp, 
				"P2\n"
				"# Created by mghdrr\n"
				"%d %d\n"
				"65536\n",
				res_c, res_r);
	}
	else {
		/* Nothing for RAW */
	}
	
	/* Compute volume boundary box */
	for (d = 0; d < 3; d++) {
		vol_limits[d].limits[0] = vol->offset[d] - 0.5 * vol->pix_spacing[d];
		vol_limits[d].limits[1] = vol_limits[d].limits[0] + vol->dim[d] * vol->pix_spacing[d];
		if (vol_limits[d].limits[0] <= vol_limits[d].limits[1]) {
			vol_limits[d].dir = 0;
			vol_limits[d].limits[0] += DRR_BOUNDARY_TOLERANCE;
			vol_limits[d].limits[1] -= DRR_BOUNDARY_TOLERANCE;
		} else {
			vol_limits[d].dir = 1;
			vol_limits[d].limits[0] -= DRR_BOUNDARY_TOLERANCE;
			vol_limits[d].limits[1] += DRR_BOUNDARY_TOLERANCE;
		}
	}
	
	/* section for parallization */

	for (r=options->image_window[0]; r<=options->image_window[1]; r++) {
		if (r % 50 == 0) printf ("Row: %4d/%d\n",r,res_r);
		vec3_copy (r_tgt, ul_room);
		vec3_scale3 (tmp, incr_r, (double) r);
		vec3_add2 (r_tgt, tmp);
		for (c=options->image_window[2]; c<=options->image_window[3]; c++) {
#if defined (ULTRA_VERBOSE)
			printf ("Row: %4d/%d col:%4d/%d\n",r,res_r,c,res_c);
#endif
			vec3_scale3 (tmp, incr_c, (double) c);
			vec3_add3 (p2, r_tgt, tmp);
			
			switch (options->interpolation) {
				case INTERPOLATION_NONE:
					value = drr_trace_ray_nointerp_2009 (vol, vol_limits, p1, p2, msd_fp);
					break;
				case INTERPOLATION_TRILINEAR_EXACT:
					value = drr_trace_ray_trilin_exact (vol, p1, p2);
					break;
				case INTERPOLATION_TRILINEAR_APPROX:
					value = drr_trace_ray_trilin_approx (vol, p1, p2);
					break;
			}
			value = value / 10;	 /* Translate from pixels to cm*gm */
			if (options->exponential_mapping) {
				value = exp(-value);
			}
			value = value * options->scale;	/* User requested scaling */
			if (options->output_format == OUTPUT_FORMAT_PFM) {
				float fv = (float) value;
				fwrite (&fv, sizeof(float), 1, pgm_fp);
				//fprintf (pgm_fp,"%g ",value);
			}
			else if (options->output_format == OUTPUT_FORMAT_PGM) {
				if (options->exponential_mapping) {
					value = value * 65536;
				} else {
					value = value * 15000;
				}
				if (value > 65536) {
					value = 65536;
				} else if (value < 0) {
					value = 0;
				}
				fprintf (pgm_fp,"%lu ", ROUND_INT(value));
			}
			else {
				/* RAW */
				short fv = (short) value;
				fwrite (&fv, sizeof(short), 1, pgm_fp);
			}
		}
		if (options->output_format == OUTPUT_FORMAT_PGM) {
			fprintf (pgm_fp,"\n");
		}
	}

	/* end section for parallization */

	fclose (pgm_fp);
	if (msd_fp) {
		fclose (msd_fp);
	}
	printf ("done.\n");
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void drr_render_volumes (Volume* vol, MGHDRR_Options* options)
{
	int a;
	
	//	double cam_ap[3] = {0.0, -1.0, 0.0};
	//	double cam_lat[3] = {-1.0, 0.0, 0.0};
	//	double* cam = cam_ap;
	//	double* cam = cam_lat;
	
	double vup[3] = {0, 0, 1};
	double tgt[3] = {0.0, 0.0, 0.0};
	double nrm[3];
	double tmp[3];
	
	/* Set source-to-axis distance */
	double sad = options->sad;
	
	/* Set source-to-image distance */
	double sid = options->sid;
	
	/* Set image resolution */
	int ires[2] = { options->image_resolution[0],
		options->image_resolution[1] };
	
	/* Set physical size of imager in mm */
	//	int isize[2] = { 300, 400 };		/* Actual resolution */
	int isize[2] = { options->image_size[0],
		options->image_size[1] };
	
	/* Set ic = image center (in pixels), and ps = pixel size (in mm)
	Note: pixels are numbered from 0 to ires-1 */
	double ic[2] = { options->image_center[0],
		options->image_center[1] };
	
	/* Set pixel size in mm */
	double ps[2] = { (double)isize[0]/(double)ires[0], 
		(double)isize[1]/(double)ires[1] };
	
	/* possible section for parallization */

	/* Loop through camera angles */
	for (a = 0; a < options->num_angles; a++) {
		double angle = a * options->angle_diff;
		double cam[3];
		char out_fn[256];
		char multispectral_fn[256];
		
		cam[0] = cos(angle);
		cam[1] = sin(angle);
		cam[2] = 0.0;
		
		printf ("Rendering DRR %d\n", a);
		
		/* Place camera at distance "sad" from the volume isocenter */
		vec3_sub3 (nrm, tgt, cam);
		vec3_normalize1 (nrm);
		vec3_scale3 (tmp, nrm, sad);
		vec3_copy (cam, tgt);
		vec3_sub2 (cam, tmp);
		
		/* Some debugging info */
#if defined (VERBOSE)
		vec_set_fmt ("%12.4g ");
		printf ("cam: ");
		vec3_print_eol (stdout, cam);
		printf ("tgt: ");
		vec3_print_eol (stdout, tgt);
		printf ("ic:  %g %g\n", ic[0], ic[1]);
#endif
		sprintf (out_fn, "%s%04d.txt", options->output_prefix, a);
		drr_write_projection_matrix (vol, cam, tgt, vup, 
									sid, ic, ps, ires, out_fn);
		if (options->output_format == OUTPUT_FORMAT_PFM) {
			sprintf (out_fn, "%s%04d.pfm", options->output_prefix, a);
		} else if (options->output_format == OUTPUT_FORMAT_PGM) {
			sprintf (out_fn, "%s%04d.pgm", options->output_prefix, a);
		} else {
			sprintf(out_fn, "%s%04d.raw", options->output_prefix, a);
		}
		sprintf (multispectral_fn, "%s%04d.msd", options->output_prefix, a);
		drr_render_volume_perspective (vol, cam, tgt, vup, 
										sid, ic, ps, 
										ires, 
										out_fn, multispectral_fn, 
										options);
	}

	/* end section for parallization */
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
Point_location drr_test_boundary (Volume_limit* vol_limit, double x)
{
	if (vol_limit->dir == 0) {
		if (x < vol_limit->limits[0]) {
			return POINTLOC_LEFT;
		} else if (x > vol_limit->limits[1]) {
			return POINTLOC_RIGHT;
		} else {
			return POINTLOC_INSIDE;
		}
	} else {
		if (x > vol_limit->limits[0]) {
			return POINTLOC_LEFT;
		} else if (x < vol_limit->limits[1]) {
			return POINTLOC_RIGHT;
		} else {
			return POINTLOC_INSIDE;
		}
	}
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
int drr_trace_init (int *ai_x, int *ai_y, int *ai_z, int *aixdir, int *aiydir, int *aizdir, double *ao_x, double *ao_y, double *ao_z, double *al_x, double *al_y, double *al_z, double *len, Volume* vol, Volume_limit vol_limits[3], double* p1, double* p2)
{
	int d;
	Point_location ploc[3][2];
	double alpha[3][2];
	double alpha_in, alpha_out;
	double ray[3];
	double ips[2][4];
	
	/* For each of 3 dimensions: x, y, and z */
	for (d = 0; d < 3; d++) {
		ploc[d][0] = drr_test_boundary (&vol_limits[d], p1[d]);
		ploc[d][1] = drr_test_boundary (&vol_limits[d], p2[d]);
		/* Reject segments which don't intersect the volume in this dimension */
		if (ploc[d][0] == POINTLOC_LEFT && ploc[d][1] == POINTLOC_LEFT) return 0;
		if (ploc[d][0] == POINTLOC_RIGHT && ploc[d][1] == POINTLOC_RIGHT) return 0;
	}
	
#if defined (ULTRA_VERBOSE)
	printf ("vol_limit[*][0] = %g %g %g\n", vol_limits[0].limits[0], vol_limits[1].limits[0], vol_limits[2].limits[0]);
	printf ("vol_limit[*][1] = %g %g %g\n", vol_limits[0].limits[1], vol_limits[1].limits[1], vol_limits[2].limits[1]);
	printf ("ploc[*][0]: %d %d %d\n", ploc[0][0], ploc[1][0], ploc[2][0]);
	printf ("ploc[*][1]: %d %d %d\n", ploc[0][1], ploc[1][1], ploc[2][1]);
#endif
	
	/* If we made it here, all three dimensions have some range of alpha
	where they intersects the volume.  However, these alphas might 
	not overlap.  We compute the alphas, then test overlapping 
	alphas to find the segment range within the volume.  */
	for (d = 0; d < 3; d++) {
		if (ploc[d][0] == POINTLOC_LEFT) {
			alpha[d][0] = (vol_limits[d].limits[0] - p1[d]) / (p2[d] - p1[d]);
		} else if (ploc[d][0] == POINTLOC_RIGHT) {
			alpha[d][0] = (p1[d] - vol_limits[d].limits[1]) / (p1[d] - p2[d]);
		} else {
			alpha[d][0] = 0.0;
		}
		if (ploc[d][1] == POINTLOC_LEFT) {
			alpha[d][1] = (vol_limits[d].limits[0] - p1[d]) / (p2[d] - p1[d]);
		} else if (ploc[d][1] == POINTLOC_RIGHT) {
			alpha[d][1] = (p1[d] - vol_limits[d].limits[1]) / (p1[d] - p2[d]);
		} else {
			alpha[d][1] = 1.0;
		}
	}
	
	/* alpha_in is the alpha where the segment enters the boundary, and 
	alpha_out is where it exits the boundary.  */
	alpha_in = alpha[0][0];
	alpha_out = alpha[0][1];
	for (d = 1; d < 3; d++) {
		if (alpha_in < alpha[d][0]) alpha_in = alpha[d][0];
		if (alpha_out > alpha[d][1]) alpha_out = alpha[d][1];
	}
#if defined (ULTRA_VERBOSE)
	printf ("alpha[*][0] = %g %g %g\n", alpha[0][0], alpha[1][0], alpha[2][0]);
	printf ("alpha[*][1] = %g %g %g\n", alpha[0][1], alpha[1][1], alpha[2][1]);
	printf ("alpha in/out = %g %g\n", alpha_in, alpha_out);
#endif
	
	/* If exit is before entrance, the segment does not intersect the volume */
	if (alpha_out - alpha_in < DRR_LEN_TOLERANCE) {
		return 0;
	}
	
	/* Create the volume intersection points */
	vec3_sub3 (ray, p2, p1);
	for (d = 0; d < 3; d++) {
		ips[0][d] = p1[d] + alpha_in * ray[d];
		ips[1][d] = p1[d] + alpha_out * ray[d];
	}
	vec3_normalize1 (ray);
	
#if defined (ULTRA_VERBOSE)
	printf ("ips[0] = %g %g %g\n", ips[0][0], ips[0][1], ips[0][2]);
	printf ("ips[1] = %g %g %g\n", ips[1][0], ips[1][1], ips[1][2]);
	printf ("ray = %g %g %g\n", ray[0], ray[1], ray[2]);
#endif

	/* We'll go from p1 to p2 */
	/* Variable notation:
	ai_x	// index of x
	aixdir  // x indices moving up or down?
	ao_x	// absolute length to next voxel crossing
	al_x	// length between voxel crossings
	*/
	drr_trace_init_loopvars_nointerp (ai_x, aixdir, ao_x, al_x, 
									 ips[0][0],
									 ray[0], 
									 vol->offset[0], 
									 vol->pix_spacing[0]);
	drr_trace_init_loopvars_nointerp (ai_y, aiydir, ao_y, al_y, 
									 ips[0][1],
									 ray[1], 
									 vol->offset[1], 
									 vol->pix_spacing[1]);
	drr_trace_init_loopvars_nointerp (ai_z, aizdir, ao_z, al_z, 
									 ips[0][2], 
									 ray[2], 
									 vol->offset[2], 
									 vol->pix_spacing[2]);
	
#if defined (ULTRA_VERBOSE)
	printf ("aix = %d aixdir = %d aox = %g alx = %g\n", *ai_x, *aixdir, *ao_x, *al_x);
	printf ("aiy = %d aiydir = %d aoy = %g aly = %g\n", *ai_y, *aiydir, *ao_y, *al_y);
	printf ("aiz = %d aizdir = %d aoz = %g alz = %g\n", *ai_z, *aizdir, *ao_z, *al_z);
#endif
	
	*len = vec3_dist(&ips[0][0],&ips[1][0]);

	return 1;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void drr_trace_init_loopvars_nointerp (int* ai, int* aidir, double* ao, double* al, double pt, double ry, double offset, double samp)
{
	
	if (ry > 0) {
		*aidir = 1;
		*ai = (int) floor ((pt - offset + 0.5 * samp) / samp);
		*ao = samp - ((pt - offset + 0.5 * samp) - (*ai) * samp);
	} else {
		*aidir = -1;
		*ai = (int) floor ((pt - offset + 0.5 * samp) / samp);
		//*ao = *ai;
		
		*ao = samp - ((*ai+1) * samp - (pt - offset + 0.5 * samp));
	}
	if (fabs(ry) > DRR_STRIDE_TOLERANCE) {
		*ao = *ao / fabs(ry);
		*al = samp / fabs(ry);
	} else {
		*ao = DRR_HUGE_DOUBLE;
		*al = DRR_HUGE_DOUBLE;
	}
}

/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
double drr_trace_ray_trilin_approx (Volume* vol, double* p1in, double* p2in)
{
	return 0.0;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
double drr_trace_ray_trilin_exact (Volume* vol, double* p1in, double* p2in)
{
	return 0.0;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
double drr_trace_ray_nointerp_2009 (Volume* vol, Volume_limit vol_limits[3], double* p1in, double* p2in, FILE* msd_fp)
{
	int ai_x, ai_y, ai_z;
	int aixdir, aiydir, aizdir;
	double ao_x, ao_y, ao_z;
	double al_x, al_y, al_z;
	double len;
	double aggr_len = 0.0;
	double accum = 0.0;
	int num_pix = 0;
	float msd_bins[MSD_NUM_BINS];
	float* img = (float*) vol->img;
	
#if defined (ULTRA_VERBOSE)
	printf ("p1in: %f %f %f\n", p1in[0], p1in[1], p1in[2]);
	printf ("p2in: %f %f %f\n", p2in[0], p2in[1], p2in[2]);
#endif

	if (!drr_trace_init (
						&ai_x,
						&ai_y,
						&ai_z,
						&aixdir, 
						&aiydir, 
						&aizdir,
						&ao_x, 
						&ao_y, 
						&ao_z,
						&al_x, 
						&al_y, 
						&al_z,
						&len,
						vol, vol_limits, p1in, p2in)) {
		return 0.0;
	}

	if (msd_fp) {
		init_multispectral (msd_bins);
	}

	/* We'll go from p1 to p2 */
	/* Variable notation:
	ai_x	// index of x
	aixdir  // x indices moving up or down?
	ao_x	// absolute length to next voxel crossing
	al_x	// length between voxel crossings
	*/

	do {
		int x = ai_z*vol->dim[0]*vol->dim[1] + ai_y*vol->dim[0]+ai_x;
		float* zz = (float*) &img[ai_z*vol->dim[0]*vol->dim[1]];
		float pix_density;
		double pix_len;
#if defined (ULTRA_VERBOSE)
		printf ("(%d %d %d) (%g,%g,%g)\n",ai_x,ai_y,ai_z,ao_x,ao_y,ao_z);
		printf ("aggr_len = %g, len = %g\n", aggr_len, len);
		fflush (stdout);
#endif
		pix_density = zz[ai_y*vol->dim[0]+ai_x];
		pix_density = img[11111111];
		if ((ao_x < ao_y) && (ao_x < ao_z)) {
			pix_len = ao_x;
			aggr_len += ao_x;
			ao_y -= ao_x;
			ao_z -= ao_x;
			ao_x = al_x;
			ai_x += aixdir;
		} else if ((ao_y < ao_z)) {
			pix_len = ao_y;
			aggr_len += ao_y;
			ao_x -= ao_y;
			ao_z -= ao_y;
			ao_y = al_y;
			ai_y += aiydir;
		} else {
			pix_len = ao_z;
			aggr_len += ao_z;
			ao_x -= ao_z;
			ao_y -= ao_z;
			ao_z = al_z;
			ai_z += aizdir;
		}
#if defined (PREPROCESS_ATTENUATION)
		accum += pix_len * pix_density;
#if defined (DEBUG_INTENSITIES)
		printf ("len: %10g dens: %10g acc: %10g\n", 
				pix_len, pix_density, accum);
#endif
#else
		accum += pix_len * attenuation_lookup (pix_density);
#endif
		if (msd_fp) {
			accumulate_multispectral (msd_bins, pix_len, pix_density);
		}
		num_pix++;
	} while (aggr_len+DRR_LEN_TOLERANCE < len);
	if (msd_fp) {
		dump_multispectral (msd_fp, msd_bins);
	}
	return accum;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void drr_write_projection_matrix (Volume* vol, double* cam, double* tgt, double* vup, double sid, double* ic, double* ps, int* ires, char* out_fn)
{
	double extrinsic[16];
	double intrinsic[12];
	double projection[12];
	const int cols = 4;
	double sad;
	
	double nrm[3];
	double vrt[3];
	double vup_tmp[3];  /* Don't overwrite vup */
	
	FILE* fp;
	
	vec_zero (extrinsic, 16);
	vec_zero (intrinsic, 12);
	
	/* Compute image coordinate sys (nrm,vup,vrt) relative to room coords.
	---------------
	nrm = tgt - cam
	vrt = nrm x vup
	vup = vrt x nrm
	---------------
	*/
	vec3_sub3 (nrm, tgt, cam);
	vec3_normalize1 (nrm);
	vec3_cross (vrt, nrm, vup);
	vec3_normalize1 (vrt);
	vec3_cross (vup_tmp, vrt, nrm);
	vec3_normalize1 (vup_tmp);
	
	/* !!! But change nrm here to -nrm */
	vec3_scale2 (nrm, -1.0);
	
	/* Build extrinsic matrix */
	vec3_copy (&extrinsic[0], vrt);
	vec3_copy (&extrinsic[4], vup_tmp);
	vec3_copy (&extrinsic[8], nrm);
	sad = vec3_len (cam);
	m_idx(extrinsic,cols,2,3) = - sad;
	m_idx(extrinsic,cols,3,3) = 1.0;
	
	/* Build intrinsic matrix */
	m_idx(intrinsic,cols,0,1) = - 1 / ps[0];
	m_idx(intrinsic,cols,1,0) = 1 / ps[1];
	m_idx(intrinsic,cols,2,2) = - 1 / sid;
	//	m_idx(intrinsic,cols,0,3) = ic[0];
	//	m_idx(intrinsic,cols,1,3) = ic[1];
	
	mat_mult_mat (projection, intrinsic,3,4, extrinsic,4,4);
	
#if defined (VERBOSE)
	printf ("Extrinsic:\n");
	matrix_print_eol (stdout, extrinsic, 4, 4);
	printf ("Intrinsic:\n");
	matrix_print_eol (stdout, intrinsic, 3, 4);
	printf ("Projection:\n");
	matrix_print_eol (stdout, projection, 3, 4);
#endif
	
	fp = fopen (out_fn, "w");
	if (!fp) {
		fprintf (stderr, "Error opening %s for write\n", out_fn);
		exit (-1);
	}
	fprintf (fp, "%18.8e %18.8e\n", ic[0], ic[1]);
	fprintf (fp,
			"%18.8e %18.8e %18.8e %18.8e\n" 
			"%18.8e %18.8e %18.8e %18.8e\n" 
			"%18.8e %18.8e %18.8e %18.8e\n", 
			projection[0], projection[1], projection[2], projection[3],
			projection[4], projection[5], projection[6], projection[7],
			projection[8], projection[9], projection[10], projection[11]
			);
	fprintf (fp, "%18.8e\n%18.8e\n", sad, sid);
	fprintf (fp, "%18.8e %18.8e %18.8e\n", nrm[0], nrm[1], nrm[2]);
	fclose (fp);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void dump_multispectral (FILE* msd_fp, float* bins)
{
	fwrite (bins, sizeof(float), MSD_NUM_BINS, msd_fp);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
cl_ulong executionTime(cl_event &event)
{
	cl_ulong start, end;
	
	clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
	clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
	
	return (end - start);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void fwrite_block (void* buf, size_t size, size_t count, FILE* fp)
{
	size_t left_to_write = count * size;
	size_t cur = 0;
	char* bufc = (char*) buf;
	
	while (left_to_write > 0) {
		size_t this_write, rc;
		
		this_write = left_to_write;
		if (this_write > WRITE_BLOCK) this_write = WRITE_BLOCK;
		rc = fwrite (&bufc[cur], 1, this_write, fp);
		if (rc != this_write) {
			fprintf (stderr, "Error writing to file."
					"rc=%zd, this_write=%zd\n",
					rc, this_write);
			return;
		}
		cur += rc;
		left_to_write -= rc;
	}
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void init_multispectral (float* bins)
{
	memset (bins, 0, MSD_NUM_BINS*sizeof(float));
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void parse_args (MGHDRR_Options* options, int argc, char* argv[])
{
	int i, rc;
	
	set_default_options (options);
	for (i = 1; i < argc; i++) {
		//printf ("ARG[%d] = %s\n", i, argv[i]);
		if (argv[i][0] != '-') break;
		if (!strcmp (argv[i], "-r")) {
			i++;
			rc = sscanf (argv[i], "%d %d", &options->image_resolution[0], &options->image_resolution[1]);
			if (rc == 1) {
				options->image_resolution[1] = options->image_resolution[0];
			} else if (rc != 2) {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-I")) {
			i++;
			options->input_file = strdup (argv[i]);
		} else if (!strcmp (argv[i], "-O")) {
			i++;
			options->output_prefix = strdup (argv[i]);
		} else if (!strcmp (argv[i], "-a")) {
			i++;
			rc = sscanf (argv[i], "%d" , &options->num_angles);
			if (rc != 1) {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-A")) {
			i++;
			rc = sscanf (argv[i], "%g" , &options->angle_diff);
			if (rc != 1) {
				print_usage ();
			}
			options->have_angle_diff = 1;
		} else if (!strcmp (argv[i], "-s")) {
			i++;
			rc = sscanf (argv[i], "%g" , &options->scale);
			if (rc != 1) {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-t")) {
			i++;
			if (!strcmp (argv[i], "pfm")) {
				options->output_format = OUTPUT_FORMAT_PFM;
			} else if (!strcmp (argv[i], "pgm")) {
				options->output_format = OUTPUT_FORMAT_PGM;
			} else if (!strcmp (argv[i], "raw")) {
				options->output_format = OUTPUT_FORMAT_RAW;
			} else {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-c")) {
			i++;
			rc = sscanf (argv[i], "%g %g", &options->image_center[0], &options->image_center[1]);
			if (rc == 1) {
				options->image_center[1] = options->image_center[0];
			} else if (rc != 2) {
				print_usage ();
			}
			options->have_image_center = 1;
		} else if (!strcmp (argv[i], "-z")) {
			i++;
			rc = sscanf (argv[i], "%g %g", &options->image_size[0], &options->image_size[1]);
			if (rc == 1) {
				options->image_size[1] = options->image_size[0];
			} else if (rc != 2) {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-g")) {
			i++;
			rc = sscanf (argv[i], "%g %g", &options->sad, &options->sid);
			if (rc != 2) {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-w")) {
			i++;
			rc = sscanf (argv[i], "%d %d %d %d",
						&options->image_window[0],
						&options->image_window[1],
						&options->image_window[2],
						&options->image_window[3]);
			if (rc == 2) {
				options->image_window[2] = options->image_window[0];
				options->image_window[3] = options->image_window[1];
			} else if (rc != 4) {
				print_usage ();
			}
			options->have_image_window = 1;
		} else if (!strcmp (argv[i], "-i")) {
			i++;
			if (!strcmp(argv[i], "exact")) {
				options->interpolation = INTERPOLATION_TRILINEAR_EXACT;
			} else if (!strcmp(argv[i], "approx")) {
				options->interpolation = INTERPOLATION_TRILINEAR_APPROX;
			} else {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-o")) {
			i++;
			rc = sscanf (argv[i], "%g %g %g" , 
						&options->isocenter[0],
						&options->isocenter[1],
						&options->isocenter[2]);
			if (rc != 3) {
				print_usage ();
			}
		} else if (!strcmp (argv[i], "-S")) {
			options->multispectral = 1;
		} else if (!strcmp (argv[i], "-e")) {
			options->exponential_mapping = 1;
		} else {
			print_usage ();
			break;
		}
	}
	
	if (!options->input_file) {
		if (i < argc) {
			options->input_file = strdup (argv[i++]);
		}
	}
	if (i < argc) {
		print_usage ();
	}
	if (!options->input_file) {
		print_usage ();
	}
	
	set_image_parms (options);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void preprocess_attenuation (Volume* vol)
{
	int i;
	float* new_img;
	float* old_img;
	
	old_img = (float*) vol->img;
	new_img = (float*) malloc (vol->npix*sizeof(float));
	
	/* section for parallization */

	for (i = 0; i < vol->npix; i++) {
		new_img[i] = attenuation_lookup (old_img[i]);
	}

	/* end section for parallization */

	vol->pix_type = PT_FLOAT;
	free (vol->img);
	vol->img = new_img;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void print_usage (void)
{
	printf ("Usage: mghdrr [options] [infile]\n"
			"Options:\n"
			" -a num				Generate num equally spaced angles\n"
			" -A ang				Difference between neighboring anges (in degrees)\n"
			" -r \"r1 r2\"		 Set output resolution (in pixels)\n"
			" -s scale			Scale the intensity of the output file\n"
			" -e					Do exponential mapping of output values\n"
			" -g \"sad sid\"		Set the sad, sid (in mm)\n"
			" -c \"c1 c2\"		 Set the image center (in pixels)\n"
			" -z \"s1 s2\"		 Set the physical size of imager (in mm)\n"
			" -w \"w1 w2 w3 w4\"  Only produce image for pixes in window (in pix)\n"
			" -t outformat		Select output format: pgm, pfm or raw\n"
			" -S					Output multispectral output files\n"
			" -i exact			Use exact trilinear interpolation\n"
			" -i approx			Use approximate trilinear interpolation\n"
			" -o \"o1 o2 o3\"	 Set isocenter position\n"
			" -I infile			Set the input file in mha format\n"
			" -O outprefix		Generate output files using the specified prefix\n"
			);
	exit (1);
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
Volume* read_mha (char* filename)
{
	int rc;
	char linebuf[LINELEN];
	Volume* vol = (Volume*) malloc (sizeof(Volume));
	int tmp;
	FILE* fp;
	
	
	fp = fopen (filename,"rb");
	if (!fp) {
		fprintf (stderr, "File %s not found\n", filename);
		return 0;
	}
	
	vol->pix_size = -1;
	vol->pix_type = PT_UNDEFINED;
	while (fgets(linebuf,LINELEN,fp)) {
		if (strcmp (linebuf, "ElementDataFile = LOCAL\n") == 0) {
			break;
		}
		if (sscanf (linebuf, "DimSize = %d %d %d",
					&vol->dim[0],
					&vol->dim[1],
					&vol->dim[2]) == 3) {
			vol->npix = vol->dim[0] * vol->dim[1] * vol->dim[2];
			continue;
		}
		if (sscanf (linebuf, "Offset = %g %g %g",
					&vol->offset[0],
					&vol->offset[1],
					&vol->offset[2]) == 3) {
			continue;
		}
		if (sscanf (linebuf, "ElementSpacing = %g %g %g",
					&vol->pix_spacing[0],
					&vol->pix_spacing[1],
					&vol->pix_spacing[2]) == 3) {
			continue;
		}
		if (sscanf (linebuf, "ElementNumberOfChannels = %d", &tmp) == 1) {
			if (vol->pix_type == PT_UNDEFINED || vol->pix_type == PT_FLOAT) {
				vol->pix_type = PT_VF_FLOAT_INTERLEAVED;
				vol->pix_size = 3*sizeof(float);
			}
			continue;
		}
		if (strcmp (linebuf, "ElementType = MET_FLOAT\n") == 0) {
			if (vol->pix_type == PT_UNDEFINED) {
				vol->pix_type = PT_FLOAT;
				vol->pix_size = sizeof(float);
			}
			continue;
		}
		if (strcmp (linebuf, "ElementType = MET_SHORT\n") == 0) {
			vol->pix_type = PT_SHORT;
			vol->pix_size = sizeof(short);
			continue;
		}
		if (strcmp (linebuf, "ElementType = MET_UCHAR\n") == 0) {
			vol->pix_type = PT_UCHAR;
			vol->pix_size = sizeof(unsigned char);
			continue;
		}
	}
	if (vol->pix_size <= 0) {
		printf ("Oops, couldn't interpret mha data type\n");
		exit (-1);
	}
	vol->img = malloc (vol->pix_size*vol->npix);
	if (!vol->img) {
		printf ("Oops, out of memory\n");
		exit (-1);
	}
	rc = fread (vol->img, vol->pix_size, vol->npix, fp);
	if (rc != vol->npix) {
		printf ("Oops, bad read from file (%d)\n", rc);
		exit (-1);
	}
	fclose (fp);
	
	return vol;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void set_default_options (MGHDRR_Options* options)
{
	options->image_resolution[0] = 128;
	options->image_resolution[1] = 128;
	options->image_size[0] = 600;
	options->image_size[1] = 600;
	options->isocenter[0] = 0.0f;
	options->isocenter[1] = 0.0f;
	options->isocenter[2] = 0.0f;
	options->have_image_center = 0;
	options->have_image_window = 0;
	options->have_angle_diff = 0;
	options->num_angles = 1;
	options->angle_diff = 1.0f;
	options->sad = 1000.0f;
	options->sid = 1630.0f;
	options->scale = 1.0f;
	options->input_file = 0;
	options->output_prefix = "out_";
	options->exponential_mapping = 0;
	options->output_format= OUTPUT_FORMAT_PFM;
	options->multispectral = 0;
	options->interpolation = INTERPOLATION_NONE;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void set_image_parms (MGHDRR_Options* options)
{
	if (!options->have_image_center) {
		options->image_center[0] = (options->image_resolution[0]-1)/2.0;
		options->image_center[1] = (options->image_resolution[1]-1)/2.0;
	}
	if (!options->have_image_window) {
		options->image_window[0] = 0;
		options->image_window[1] = options->image_resolution[0] - 1;
		options->image_window[2] = 0;
		options->image_window[3] = options->image_resolution[1] - 1;
	}
	if (options->have_angle_diff) {
		options->angle_diff *= (float) (M_TWOPI / 360.0);
	} else {
		options->angle_diff = M_TWOPI / options->num_angles;
	}
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void set_isocenter (Volume* vol, MGHDRR_Options* options)
{
	vol->offset[0] -= options->isocenter[0];
	vol->offset[1] -= options->isocenter[1];
	vol->offset[2] -= options->isocenter[2];
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void volume_convert_to_float (Volume* ref)
{
	switch (ref->pix_type) {
		case PT_UCHAR:
			fprintf (stderr, "Sorry, UCHAR to FLOAT is not implemented\n");
			exit (-1);
			break;
		case PT_SHORT:
		{
			int v;
			short* old_img;
			float* new_img;
			
			old_img = (short*) ref->img;
			new_img = (float*) malloc (sizeof(float) * ref->npix);
			if (!new_img) {
				fprintf (stderr, "Memory allocation failed.\n");
				exit(1);
			}
			for (v = 0; v < ref->npix; v++) {
				new_img[v] = (float) old_img[v];
			}
			ref->pix_size = sizeof(float);
			ref->pix_type = PT_FLOAT;
			ref->img = (void*) new_img;
			free (old_img);
		}
			break;
		case PT_FLOAT:
			/* Nothing to do */
			break;
		case PT_UINT32:
		case PT_VF_FLOAT_INTERLEAVED:
		case PT_VF_FLOAT_PLANAR:
		default:
			/* Can't convert this */
			fprintf (stderr, "Sorry, unsupported conversion to FLOAT\n");
			exit (-1);
			break;
	}
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
int main(int argc, char* argv[])
{
	// Initialize variables
	float total_runtime;

	// Set logfile name and start logs
	shrSetLogFileName ("drr_opencl.txt");

	// Output command call
	shrLog("\n\n");
	for (int i = 0; i < argc; i++)
		shrLog("%s ", argv[i]);
	shrLog("\n\nStarting DRR_OPENCL...\n\n"); 

	/**************************************************************** 
	* STEP 1: Setup OpenCL														* 
	****************************************************************/

	// Get the OpenCL platform
	cl_platform_id platform;
	error = oclGetPlatformID(&platform);
	oclCheckError(error, CL_SUCCESS);

	// get devices of type GPU
	error = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 0, NULL, &device_count);
	oclCheckError(error, CL_SUCCESS);

	// makes sure using no more than the maximum number of GPUs
	if (device_count > MAX_GPU_COUNT)
		device_count = MAX_GPU_COUNT;

	devices = (cl_device_id *)malloc(device_count * sizeof(cl_device_id) );
	error = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, device_count, devices, NULL);
	oclCheckError(error, CL_SUCCESS);

	// create context
	cl_context_properties properties[] = {CL_CONTEXT_PLATFORM, (cl_context_properties)platform, 0};
	context = clCreateContext(properties, device_count, devices, NULL, NULL, &error);
	oclCheckError(error, CL_SUCCESS);

	shrLog("Using %d device(s):\n", device_count);

	for (cl_uint i = 0; i < device_count; i++) {
		// device info
		device = oclGetDev(context, i);
		clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(device_name[i]), device_name[i], NULL);
		oclCheckError(error, CL_SUCCESS);
		shrLog("\tDevice %d: %s\n", i, device_name);

		// command queue
		command_queue[i] = clCreateCommandQueue(context, device, CL_QUEUE_PROFILING_ENABLE, &error);
		oclCheckError(error, CL_SUCCESS);
	}

	// Program Setup
	size_t program_length;
	char* source_path = shrFindFilePath("drr_opencl.cl", argv[0]);
	oclCheckError(source_path != NULL, shrTRUE);
	char *source = oclLoadProgSource(source_path, "", &program_length);
	oclCheckError(source != NULL, shrTRUE);

	// Create the program
	program = clCreateProgramWithSource(context, 1, (const char **)&source, &program_length, &error);
	oclCheckError(error, CL_SUCCESS);
	
	// Build the program
	error = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
	if (error != CL_SUCCESS)
	{
		// write out standard error, Build Log and PTX, then return error
		shrLogEx(LOGBOTH | ERRORMSG, error, STDERROR);
		oclLogBuildInfo(program, oclGetFirstDev(context));
		 oclLogPtx(program, oclGetFirstDev(context), "drr_opencl.ptx");
		 return(EXIT_FAILURE); 
	}

	for (cl_uint i = 0; i < device_count; i++)
		clFinish(command_queue[i]);
	/****************************************************************/

	Volume* vol;
	MGHDRR_Options options;
	
	parse_args (&options, argc, argv);
	
	vol = read_mha (options.input_file);
	volume_convert_to_float (vol);
	set_isocenter (vol, &options);

	total_runtime = preprocess_attenuation_and_drr_render_volumes(vol, &options);

	/*************************************
	 * STEP 5: Cleanup OpenCL and finish  *
	 *************************************/
	// Cleanup opencl
	clReleaseProgram(program);
	for (cl_uint i = 0; i < device_count; i++)
		clReleaseCommandQueue(command_queue[i]);

	clReleaseContext(context);

	shrLog("\n\nDone DRR_OPENCL...\n\n");

	shrLog("Total OpenCL run time: %f s\n", total_runtime);

	return 0;
}


/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
float preprocess_attenuation_and_drr_render_volumes(Volume* vol, MGHDRR_Options* options)
{
	// Create variables for memory timing
	cl_event vol_event[MAX_GPU_COUNT];
	cl_ulong drr_total[MAX_GPU_COUNT], img_total[MAX_GPU_COUNT], multispectral_total[MAX_GPU_COUNT], valid_total[MAX_GPU_COUNT], vol_total[MAX_GPU_COUNT];
	for (cl_uint i = 0; i < device_count; i++) {
		drr_total[i] = 0;
		img_total[i] = 0;
		multispectral_total[i] = 0;
		valid_total[i] = 0;
		vol_total[i] = 0;
	}

	// Calculate dynamic size of memory buffers
	int image_height = options->image_window[1] - options->image_window[0] + 1;
	int image_width = options->image_window[3] - options->image_window[2] + 1;
	int float3_size = 3 * sizeof(float);
	int vol_size = (vol->dim[0] * vol->dim[1] * vol->dim[2]) * sizeof(float);
	int img_size = image_height * image_width * sizeof(float);
	int multispectral_size = img_size * MSD_NUM_BINS;
	int valid_size = image_height * image_width * sizeof(bool);

	int4 pixels_per_device[MAX_GPU_COUNT];
	int2 ppd;
	int extra;
	int offset[MAX_GPU_COUNT];
	int img_size_device[MAX_GPU_COUNT];
	int multispectral_size_device[MAX_GPU_COUNT];
	int valid_size_device[MAX_GPU_COUNT];

	// Calculate number of pixels per device
	ppd.x = image_width;
	ppd.y = image_height / device_count;

	// Calculate any extra pixels allotted to last device
	extra = image_height % ppd.y;

	// Allocate pixels to each device 
	for (cl_uint i = 0; i < device_count; i++) {
		pixels_per_device[i].x = ppd.x;
		pixels_per_device[i].y = ppd.y;
		offset[i] = i * ppd.y;
	}

	// Allocate any extra pixels to the last device
	pixels_per_device[device_count - 1].y += extra;

	for (cl_uint i = 0; i < device_count; i++) {
		pixels_per_device[i].z = pixels_per_device[i].x * pixels_per_device[i].y;
		img_size_device[i] = pixels_per_device[i].z * sizeof(float);
		multispectral_size_device[i] = pixels_per_device[i].z * sizeof(float) * MSD_NUM_BINS;
		valid_size_device[i] = pixels_per_device[i].z * sizeof(bool);
	}

	// Declare all global memory buffers
	cl_mem g_dev_vol[MAX_GPU_COUNT];
	cl_mem g_dev_img[MAX_GPU_COUNT];
	cl_mem g_dev_multispectral[MAX_GPU_COUNT];
	cl_mem g_dev_valid[MAX_GPU_COUNT];
	
	// Delcare all constant memory buffers
	cl_mem c_vol_dim[MAX_GPU_COUNT];
	cl_mem c_img_dim[MAX_GPU_COUNT];
	cl_mem c_offset[MAX_GPU_COUNT];
	cl_mem c_pix_spacing[MAX_GPU_COUNT];
	cl_mem c_vol_limits[MAX_GPU_COUNT];
	cl_mem c_p1[MAX_GPU_COUNT];
	cl_mem c_ul_room[MAX_GPU_COUNT];
	cl_mem c_incr_r[MAX_GPU_COUNT];
	cl_mem c_incr_c[MAX_GPU_COUNT];
	cl_mem c_pixel_device[MAX_GPU_COUNT];

	// Declare all kernels
	cl_kernel drr_kernel[MAX_GPU_COUNT];

	for (cl_uint i = 0; i < device_count; i++) {
		// Allocate global memory on all devices
		g_dev_vol[i] = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, vol_size, NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		g_dev_img[i] = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, img_size_device[i], NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		g_dev_multispectral[i] = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, multispectral_size_device[i], NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		g_dev_valid[i] = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, valid_size_device[i], NULL, &error);
		oclCheckError(error, CL_SUCCESS);

		// Allocate constant memory on all devices
		c_vol_dim[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, 3 * sizeof(int), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_img_dim[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(int2), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_offset[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, float3_size, NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_pix_spacing[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, float3_size, NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_vol_limits[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, 3 * sizeof(Volume_limit), NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_p1[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, float3_size, NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_ul_room[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, float3_size, NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_incr_r[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, float3_size, NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_incr_c[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, float3_size, NULL, &error);
		oclCheckError(error, CL_SUCCESS);
		c_pixel_device[i] = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_ALLOC_HOST_PTR, sizeof(int4), NULL, &error);
		oclCheckError(error, CL_SUCCESS);

		// Create the drr kernel on all devices
		drr_kernel[i] = clCreateKernel(program, "kernel_drr", &error);
		oclCheckError(error, CL_SUCCESS);
	}

	// Wait for all queues to finish
	for (cl_uint i = 0; i < device_count; i++)
		clFinish(command_queue[i]);

	// Calculate the drr runtime environment
	size_t drr_local_work_size[MAX_GPU_COUNT][2];
	size_t drr_global_work_size[MAX_GPU_COUNT][2];
	
	for (cl_uint i = 0; i < device_count; i++) {
		drr_local_work_size[i][0] = 128;
		drr_local_work_size[i][1] = 1;
		drr_global_work_size[i][0] = shrRoundUp((int)drr_local_work_size[i][0], pixels_per_device[i].x);
		drr_global_work_size[i][1] = shrRoundUp((int)drr_local_work_size[i][1], pixels_per_device[i].y);
	}

	// Copy memory from host to device
	for (cl_uint i = 0; i < device_count; i++) {
		error = clEnqueueWriteBuffer(command_queue[i], g_dev_vol[i], CL_FALSE, 0, vol_size, vol->img, 0, NULL, &vol_event[i]);
		oclCheckError(error, CL_SUCCESS);
	}

	// Waits for volume to finish copying
	clWaitForEvents(device_count, vol_event);

	// Count host to device time
	for (cl_uint i = 0; i < device_count; i++)
		vol_total[i] += executionTime(vol_event[i]);

	#if defined (PREPROCESS_ATTENUATION)
		// Create variables for preprocess kernel and memory timing
		cl_event preprocess_event[MAX_GPU_COUNT];
		cl_ulong preprocess_total[MAX_GPU_COUNT];
		for (cl_uint i = 0; i < device_count; i++)
			preprocess_total[i] = 0;

		if (vol->pix_type != PT_FLOAT) {
			float *old_img = (float*)vol->img;
			float *new_img = (float*)malloc(vol->npix * sizeof(float));

			for (int i = 0; i < vol->npix; i++) {
				new_img[i] = old_img[i];
			}

			vol->pix_type = PT_FLOAT;
			free(vol->img);
			vol->img = new_img;

			// Update global memory from host to device
			for (cl_uint i = 0; i < device_count; i++) {
				error = clEnqueueWriteBuffer(command_queue[i], g_dev_vol[i], CL_FALSE, 0, vol_size, vol->img, 0, NULL, &vol_event[i]);
				oclCheckError(error, CL_SUCCESS);
			}

			// Waits for volume to finish copying
			clWaitForEvents(device_count, vol_event);

			// Count host to device time
			for (cl_uint i = 0; i < device_count; i++)
				vol_total[i] += executionTime(vol_event[i]);
		}

		// Start working
		shrLog("\nPerforming preprocess attenuation...\n\n");

		// Creates the kernel object
		cl_kernel preprocess_kernel[MAX_GPU_COUNT];
		for (cl_uint i = 0; i < device_count; i++) {
			preprocess_kernel[i] = clCreateKernel(program, "preprocess_attenuation_cl", &error);
			oclCheckError(error, CL_SUCCESS);
		}

		// Sets Kernel Arguments
		for (cl_uint i = 0; i < device_count; i++) {
			error = clSetKernelArg(preprocess_kernel[i], 0, sizeof(cl_mem), (void *) &g_dev_vol[i]);
			error |= clSetKernelArg(preprocess_kernel[i], 1, sizeof(int), &vol->npix);
			oclCheckError(error, CL_SUCCESS);
		}

		for (cl_uint i = 0; i < device_count; i++)
			clFinish(command_queue[i]);

		// Enqueue kernel and execute
		size_t preprocess_local_work_size[MAX_GPU_COUNT];
		size_t preprocess_global_work_size[MAX_GPU_COUNT];
		for (cl_uint i=0; i < device_count; i++) {
			preprocess_local_work_size[i] = 512;
			preprocess_global_work_size[i] = shrRoundUp((int)preprocess_local_work_size[i], vol->npix);
		}

		for (cl_uint i = 0; i < device_count; i++) {
			error = clEnqueueNDRangeKernel(command_queue[i], preprocess_kernel[i], 1, NULL, &preprocess_global_work_size[i], &preprocess_local_work_size[i], 0, NULL, &preprocess_event[i]);
			oclCheckError(error, CL_SUCCESS);
		}

		// Waits for queue to finish
		clWaitForEvents(device_count, preprocess_event);

		// Count kernel time
		for (cl_uint i = 0; i < device_count; i++)
			preprocess_total[i] += executionTime(preprocess_event[i]);

		shrLog("Done preprocessing ...\n\n");

		// Release kernels and memory
		for (cl_uint i = 0; i < device_count; i++) {
			clReleaseKernel(preprocess_kernel[i]);
		}

	#endif

	shrLog("Rendering DRR images\n\n");

	double vup[3] = {0, 0, 1};
	double tgt[3] = {0.0, 0.0, 0.0};
	double nrm[3];
	double tmp[3];
	
	// Set source-to-axis distance
	double sad = options->sad;
	
	// Set source-to-image distance
	double sid = options->sid;
	
	// Set image resolution
	int ires[2] = { options->image_resolution[0], options->image_resolution[1] };
	
	// Set physical size of imager in mm
	// int isize[2] = { 300, 400 };		/* Actual resolution */
	int isize[2] = { options->image_size[0], options->image_size[1] };
	
	// Set ic = image center (in pixels), and ps = pixel size (in mm)
	// Note: pixels are numbered from 0 to ires-1
	double ic[2] = { options->image_center[0], options->image_center[1] };
	
	// Set pixel size in mm
	double ps[2] = { (double)isize[0] / (double)ires[0], (double)isize[1] / (double)ires[1] };

	// Loop through camera angles
	for (int a = 0; a < options->num_angles; a++) {
		double angle = a * options->angle_diff;
		double cam[3];
		char out_fn[256];
		char multispectral_fn[256];
		
		cam[0] = cos(angle);
		cam[1] = sin(angle);
		cam[2] = 0.0;
		
		shrLog("Rendering DRR %d\n", a);
		
		/* Place camera at distance "sad" from the volume isocenter */
		vec3_sub3(nrm, tgt, cam);
		vec3_normalize1(nrm);
		vec3_scale3(tmp, nrm, sad);
		vec3_copy(cam, tgt);
		vec3_sub2(cam, tmp);
		
		sprintf(out_fn, "%s%04d.txt", options->output_prefix, a);
		drr_write_projection_matrix(vol, cam, tgt, vup, sid, ic, ps, ires, out_fn);
		if (options->output_format == OUTPUT_FORMAT_PFM) {
			sprintf(out_fn, "%s%04d.pfm", options->output_prefix, a);
		} else if (options->output_format == OUTPUT_FORMAT_PGM) {
			sprintf(out_fn, "%s%04d.pgm", options->output_prefix, a);
		} else {
			sprintf(out_fn, "%s%04d.raw", options->output_prefix, a);
		}
		sprintf(multispectral_fn, "%s%04d.msd", options->output_prefix, a);
		drr_render_volume_perspective_cl(vol, cam, tgt, vup, sid, ic, ps, ires, out_fn, multispectral_fn, options, img_size, multispectral_size, valid_size, float3_size, g_dev_vol, g_dev_img, g_dev_multispectral, g_dev_valid, c_vol_dim, c_img_dim, c_offset, c_pix_spacing, c_vol_limits, c_p1, c_ul_room, c_incr_r, c_incr_c, c_pixel_device, drr_kernel, drr_total, img_total, multispectral_total, valid_total, drr_global_work_size, drr_local_work_size, pixels_per_device, offset, img_size_device, multispectral_size_device, valid_size_device);
		shrLog("Done rendering...\n\n", a);
	}

	for (cl_uint i = 0; i < device_count; i++) {
		// Release kernels
		clReleaseKernel(drr_kernel[i]);

		// Release constant memory buffers
		clReleaseMemObject(c_pixel_device[i]);
		clReleaseMemObject(c_incr_c[i]);
		clReleaseMemObject(c_incr_r[i]);
		clReleaseMemObject(c_ul_room[i]);
		clReleaseMemObject(c_p1[i]);
		clReleaseMemObject(c_vol_limits[i]);
		clReleaseMemObject(c_pix_spacing[i]);
		clReleaseMemObject(c_offset[i]);
		clReleaseMemObject(c_img_dim[i]);
		clReleaseMemObject(c_vol_dim[i]);

		// Release global memory buffers
		clReleaseMemObject(g_dev_valid[i]);
		clReleaseMemObject(g_dev_multispectral[i]);
		clReleaseMemObject(g_dev_img[i]);
		clReleaseMemObject(g_dev_vol[i]);
	}

	// Perform timing
	float overall_runtime = 0;
	for (cl_uint i = 0; i < device_count; i++) {
		float drr_kernel_runtime = drr_total[i] * 1.0e-6f;
		float preprocess_kernel_runtime = preprocess_total[i] * 1.0e-6f;
		float img_copy_runtime = img_total[i] * 1.0e-6f;
		float vol_copy_runtime = vol_total[i] * 1.0e-6f;
		float total_runtime = (drr_kernel_runtime + preprocess_kernel_runtime + img_copy_runtime + vol_copy_runtime) * 1.0e-3f;
		overall_runtime += total_runtime;

		shrLog("Device %d: %s\n", i, device_name);
		shrLog("\tDRR Kernel run time: %f ms\n", drr_kernel_runtime);
		shrLog("\tHu Kernel run time: %f ms\n", preprocess_kernel_runtime);
		shrLog("\tImage host/device & device/host copy run time: %f ms\n", img_copy_runtime);
		shrLog("\tVolume host to device copy run time: %f ms\n\n", vol_copy_runtime);
	}

	return overall_runtime;
}

/*
 * Summary:		
 * Parameters:	
 * Return:		
 */
void drr_render_volume_perspective_cl(Volume* vol, double* cam, double* tgt, double* vup, double sid, double* ic, double* ps, int* ires, char* image_fn, char* multispectral_fn,  MGHDRR_Options* options, int img_size, int multispectral_size, int valid_size, int float3_size, cl_mem *g_dev_vol, cl_mem *g_dev_img, cl_mem *g_dev_multispectral, cl_mem *g_dev_valid, cl_mem *c_vol_dim, cl_mem *c_img_dim, cl_mem *c_offset, cl_mem *c_pix_spacing, cl_mem *c_vol_limits, cl_mem *c_p1, cl_mem *c_ul_room, cl_mem *c_incr_r, cl_mem *c_incr_c, cl_mem *c_pixel_device, cl_kernel *drr_kernel, cl_ulong *drr_total, cl_ulong *img_total, cl_ulong *multispectral_total, cl_ulong *valid_total, size_t drr_global_work_size[MAX_GPU_COUNT][2], size_t drr_local_work_size[MAX_GPU_COUNT][2], int4 *pixels_per_device, int *offset, int *img_size_device, int *multispectral_size_device, int *valid_size_device)
{
	int d;
	double nrm[3];
	double vrt[3];
	double p1[3];
	int res_r = options->image_window[1] - options->image_window[0] + 1;
	int res_c = options->image_window[3] - options->image_window[2] + 1;
	double pp_r = - ps[0];
	double pp_c = ps[1];
	double ic_room[3];
	double ul_room[3];
	double incr_r[3];
	double incr_c[3];
	double tmp[3];
	Volume_limit vol_limits[3];
	
	FILE *pgm_fp, *msd_fp;
	int r, c;
	float value;
	
	/* Compute image coordinate sys (nrm,vup,vrt) relative to room coords.
	---------------
	nrm = tgt - cam
	vrt = nrm x vup
	vup = vrt x nrm
	---------------
	*/
	vec3_sub3(nrm, tgt, cam);
	vec3_normalize1(nrm);
	vec3_cross(vrt, nrm, vup);
	vec3_normalize1(vrt);
	vec3_cross(vup, vrt, nrm);
	vec3_normalize1(vup);
	
	/* Compute position of image center in room coordinates */
	vec3_scale3(tmp, nrm, sid);
	vec3_add3(ic_room, cam, tmp);
	#if defined (VERBOSE)
		printf ("icr: ");
		vec3_print_eol (stdout, ic_room);
	#endif
	
	/* Compute base and increments for pixels. Base is at upper left, 
	and increment with increasing rows and columns. */
	vec3_scale3(incr_r, vup, pp_r);
	vec3_scale3(incr_c, vrt, pp_c);
	vec3_copy(ul_room, ic_room);
	vec3_scale3(tmp, incr_r, -ic[0]);
	vec3_add2(ul_room, tmp);
	vec3_scale3(tmp, incr_c, -ic[1]);
	vec3_add2(ul_room, tmp);
	
	/* drr_trace_ray uses p1 & p2, p1 is the camera, p2 is in the 
	direction of the ray */
	vec3_copy(p1, cam);
	
	pgm_fp = fopen(image_fn,"wb");
	if (!pgm_fp) {
		fprintf(stderr, "Error opening %s for write\n", image_fn);
		exit(-1);
	}

	if (options->multispectral) {
		msd_fp = fopen(multispectral_fn,"wb");
		if (!msd_fp) {
			fprintf(stderr, "Error opening %s for write\n", multispectral_fn);
			exit(-1);
		}
	} else {
		msd_fp = 0;
	}
	
	if (options->output_format == OUTPUT_FORMAT_PFM) {
		fprintf(pgm_fp, 
				"Pf\n"
				"%d %d\n"
				"-1\n",
				res_c, res_r);
	} else if (options->output_format == OUTPUT_FORMAT_PGM) {
		fprintf(pgm_fp, 
				"P2\n"
				"# Created by mghdrr\n"
				"%d %d\n"
				"65536\n",
				res_c, res_r);
	} else {
		/* Nothing for RAW */
	}
	
	if (options->multispectral) {
		msd_fp = fopen(multispectral_fn,"wb");
		if (!msd_fp) {
			fprintf(stderr, "Error opening %s for write\n", multispectral_fn);
			exit(-1);
		}
	} else {
		msd_fp = 0;
	}
	
	/* Compute volume boundary box */
	for (d = 0; d < 3; d++) {
		vol_limits[d].limits[0] = vol->offset[d] - 0.5 * vol->pix_spacing[d];
		vol_limits[d].limits[1] = vol_limits[d].limits[0] + vol->dim[d] * vol->pix_spacing[d];
		if (vol_limits[d].limits[0] <= vol_limits[d].limits[1]) {
			vol_limits[d].dir = 0;
			vol_limits[d].limits[0] += DRR_BOUNDARY_TOLERANCE;
			vol_limits[d].limits[1] -= DRR_BOUNDARY_TOLERANCE;
		} else {
			vol_limits[d].dir = 1;
			vol_limits[d].limits[0] -= DRR_BOUNDARY_TOLERANCE;
			vol_limits[d].limits[1] += DRR_BOUNDARY_TOLERANCE;
		}
	}

	// Create variables for kernel and memory timing
	cl_event drr_event[MAX_GPU_COUNT], img_event[MAX_GPU_COUNT], multispectral_event[MAX_GPU_COUNT], valid_event[MAX_GPU_COUNT];

	// Start working
	shrLog("Processing image...\n");

	float *img = (float*)malloc(img_size);
	float *multispectral = (float*)malloc(multispectral_size);
	bool *valid = (bool*)malloc(valid_size);
	int2 img_dim = {res_r, res_c};

	// Convert all doubles to floats
	float p1_f[3];
	float ul_room_f[3];
	float incr_r_f[3];
	float incr_c_f[3];
	Volume_limit_f vol_limits_f[3];
	for (int i = 0; i < 3; i++) {
		p1_f[i] = (float)p1[i];
		ul_room_f[i] = (float)ul_room[i];
		incr_r_f[i] = (float)incr_r[i];
		incr_c_f[i] = (float)incr_c[i];
		vol_limits_f[i].limits[0] = (float)vol_limits[i].limits[0];
		vol_limits_f[i].limits[1] = (float)vol_limits[i].limits[1];
		vol_limits_f[i].dir = vol_limits[i].dir;
	}
	int preprocess_attenuation = PREPROCESS_ATTENUATION;

	// Copy constant memory from host to device
	for (cl_uint i = 0; i < device_count; i++) {
		error = clEnqueueWriteBuffer(command_queue[i], g_dev_img[i], CL_FALSE, 0, img_size_device[i], (float*)img + offset[i], 0, NULL, &img_event[i]);
		error |= clEnqueueWriteBuffer(command_queue[i], g_dev_multispectral[i], CL_FALSE, 0, multispectral_size_device[i], (float*)multispectral + (offset[i] * MSD_NUM_BINS), 0, NULL, &multispectral_event[i]);
		error |= clEnqueueWriteBuffer(command_queue[i], g_dev_valid[i], CL_FALSE, 0, valid_size_device[i], (bool*)valid + offset[i], 0, NULL, &valid_event[i]);
		error |= clEnqueueWriteBuffer(command_queue[i], c_vol_dim[i], CL_FALSE, 0, 3 * sizeof(int), &vol->dim, 0, NULL, NULL);
		error |= clEnqueueWriteBuffer(command_queue[i], c_img_dim[i], CL_FALSE, 0, sizeof(int2), &img_dim, 0, NULL, NULL);
		error |= clEnqueueWriteBuffer(command_queue[i], c_offset[i], CL_FALSE, 0, float3_size, &vol->offset, 0, NULL, NULL);
		error |= clEnqueueWriteBuffer(command_queue[i], c_pix_spacing[i], CL_FALSE, 0, float3_size, &vol->pix_spacing, 0, NULL, NULL);
		error |= clEnqueueWriteBuffer(command_queue[i], c_vol_limits[i], CL_FALSE, 0, 3 * sizeof(Volume_limit), vol_limits_f, 0, NULL, NULL);
		error |= clEnqueueWriteBuffer(command_queue[i], c_p1[i], CL_FALSE, 0, float3_size, p1_f, 0, NULL, NULL);
		error |= clEnqueueWriteBuffer(command_queue[i], c_ul_room[i], CL_FALSE, 0, float3_size, &ul_room_f, 0, NULL, NULL);
		error |= clEnqueueWriteBuffer(command_queue[i], c_incr_r[i], CL_FALSE, 0, float3_size, incr_r_f, 0, NULL, NULL);
		error |= clEnqueueWriteBuffer(command_queue[i], c_incr_c[i], CL_FALSE, 0, float3_size, incr_c_f, 0, NULL, NULL);
		error |= clEnqueueWriteBuffer(command_queue[i], c_pixel_device[i], CL_FALSE, 0, sizeof(int4), &pixels_per_device[i], 0, NULL, NULL);
		oclCheckError(error, CL_SUCCESS);
	}

	// Wait for all queues to finish
	for (cl_uint i = 0; i < device_count; i++)
		clFinish(command_queue[i]);

	// Count device time
	for (cl_uint i = 0; i < device_count; i++) {
		*img_total += executionTime(img_event[i]);
		*multispectral_total += executionTime(multispectral_event[i]);
		*valid_total += executionTime(valid_event[i]);
	}

	// Invoke kernel
	// Sets Kernel Arguments
	for (cl_uint i = 0; i < device_count; i++) {
		error |= clSetKernelArg(drr_kernel[i], 0, sizeof(cl_mem), (void *) &g_dev_vol[i]);
		error |= clSetKernelArg(drr_kernel[i], 1, sizeof(cl_mem), (void *) &g_dev_img[i]);
		error |= clSetKernelArg(drr_kernel[i], 2, sizeof(cl_mem), (void *) &g_dev_multispectral[i]);
		error |= clSetKernelArg(drr_kernel[i], 3, sizeof(cl_mem), (void *) &g_dev_valid[i]);
		error |= clSetKernelArg(drr_kernel[i], 4, sizeof(cl_mem), (void *) &c_vol_dim[i]);
		error |= clSetKernelArg(drr_kernel[i], 5, sizeof(cl_mem), (void *) &c_img_dim[i]);
		error |= clSetKernelArg(drr_kernel[i], 6, sizeof(cl_mem), (void *) &c_offset[i]);
		error |= clSetKernelArg(drr_kernel[i], 7, sizeof(cl_mem), (void *) &c_pix_spacing[i]);
		error |= clSetKernelArg(drr_kernel[i], 8, sizeof(cl_mem), (void *) &c_vol_limits[i]);
		error |= clSetKernelArg(drr_kernel[i], 9, sizeof(cl_mem), (void *) &c_p1[i]);
		error |= clSetKernelArg(drr_kernel[i], 10, sizeof(cl_mem), (void *) &c_ul_room[i]);
		error |= clSetKernelArg(drr_kernel[i], 11, sizeof(cl_mem), (void *) &c_incr_r[i]);
		error |= clSetKernelArg(drr_kernel[i], 12, sizeof(cl_mem), (void *) &c_incr_c[i]);
		error |= clSetKernelArg(drr_kernel[i], 13, sizeof(cl_mem), (void *) &c_pixel_device[i]);
		error |= clSetKernelArg(drr_kernel[i], 14, sizeof(float), &options->scale);
		error |= clSetKernelArg(drr_kernel[i], 15, sizeof(int), &options->output_format);
		error |= clSetKernelArg(drr_kernel[i], 16, sizeof(int), &msd_fp);
		error |= clSetKernelArg(drr_kernel[i], 17, sizeof(int), &preprocess_attenuation);
		error |= clSetKernelArg(drr_kernel[i], 18, sizeof(int), &options->exponential_mapping);
		error |= clSetKernelArg(drr_kernel[i], 19, sizeof(int), &offset[i]);
		oclCheckError(error, CL_SUCCESS);
	}
	
	// Wait for all queues to finish
	for (cl_uint i = 0; i < device_count; i++)
		clFinish(command_queue[i]);

	// Invoke all kernels
	for (cl_uint i = 0; i < device_count; i++) {
		error = clEnqueueNDRangeKernel(command_queue[i], drr_kernel[i], 2, NULL, drr_global_work_size[i], drr_local_work_size[i], 0, NULL, &drr_event[i]);
		oclCheckError(error, CL_SUCCESS);
	}

	// Wait for fdk kernel to finish
	clWaitForEvents(device_count, drr_event);

	// Count kernel time
	for (cl_uint i = 0; i < device_count; i++)
		*drr_total += executionTime(drr_event[i]);

	// Copy img/multispectral from device to host
	for (cl_uint i = 0; i < device_count; i++) {
		error = clEnqueueReadBuffer(command_queue[i], g_dev_img[i], CL_FALSE, 0, img_size_device[i], (float*)img + offset[i], 0, NULL, &img_event[i]);
		oclCheckError(error, CL_SUCCESS);
		error = clEnqueueReadBuffer(command_queue[i], g_dev_multispectral[i], CL_FALSE, 0, multispectral_size_device[i], (float*)multispectral + (offset[i] * MSD_NUM_BINS), 0, NULL, &multispectral_event[i]);
		oclCheckError(error, CL_SUCCESS);
		error = clEnqueueReadBuffer(command_queue[i], g_dev_valid[i], CL_FALSE, 0, valid_size_device[i], (bool*)valid + offset[i], 0, NULL, &valid_event[i]);
		oclCheckError(error, CL_SUCCESS);
	}

	for (cl_uint i = 0; i < device_count; i++)
		clFinish(command_queue[i]);

	shrLog("Done processing image...\n");

	// Count device time
	for (cl_uint i = 0; i < device_count; i++) {
		*img_total += executionTime(img_event[i]);
		*multispectral_total += executionTime(multispectral_event[i]);
		*valid_total += executionTime(valid_event[i]);
	}

	for (r = options->image_window[0]; r <= options->image_window[1]; r++) {
		for (c = options->image_window[2]; c <= options->image_window[3]; c++) {
			value = img[r * res_c + c];
			if (options->output_format == OUTPUT_FORMAT_PFM) {
				fwrite(&value, sizeof(float), 1, pgm_fp);
			} else if (options->output_format == OUTPUT_FORMAT_PGM) {
				fprintf(pgm_fp, "%lu ", ROUND_INT(value));
			} else {
				/* RAW */
				short fv = (short)value;
				fwrite(&fv, sizeof(short), 1, pgm_fp);
			}
			if (msd_fp) {
				if (valid[r * res_c + c])
					fwrite(&multispectral[(r * res_c + c) * MSD_NUM_BINS], sizeof(float), MSD_NUM_BINS, msd_fp);
			}
		}
		if (options->output_format == OUTPUT_FORMAT_PGM) {
			fprintf(pgm_fp, "\n");
		}
	}

	free(img);
	free(multispectral);
	free(valid);

	fclose(pgm_fp);
	if (msd_fp) {
		fclose(msd_fp);
	}
}
