#ifndef _DRR_OPENCL_H_
#define _DRR_OPENCL_H_

#include <stdio.h>
#include <stdlib.h>
#include <cstdlib>
#include <math.h>
#include <string.h>
#include <fstream>
#include <oclUtils.h>
#include "mathutil.h"

struct int2 {
	int x;
	int y;
};

struct int3 {
	int x;
	int y;
	int z;
};

struct int4 {
	int x;
	int y;
	int z;
	int w;
};

struct float2 {
	float x;
	float y;
};

struct float3 {
	float x;
	float y;
	float z;
};

struct float4 {
	float x;
	float y;
	float z;
	float w;
};

#define MAX_GPU_COUNT 8

#define DRR_BOUNDARY_TOLERANCE 1e-6
#define DRR_HUGE_DOUBLE 1e10
#define DRR_LEN_TOLERANCE 1e-6
#define DRR_PLANE_RAY_TOLERANCE 1e-8
#define DRR_STRIDE_TOLERANCE 1e-10
#define DRR_TOPLANE_TOLERANCE 1e-7

#define INTERPOLATION_NONE 0
#define INTERPOLATION_TRILINEAR_APPROX 2
#define INTERPOLATION_TRILINEAR_EXACT 1

#define MSD_NUM_BINS 60

#define OPTION_ANGLE_DIFF 'A'
#define OPTION_ANGLE_DIFF_STRING "angle-diff"
#define OPTION_EXPONENTIAL 'e'
#define OPTION_EXPONENTIAL_STRING "exponential"
#define OPTION_IMAGE_CENTER 'c'
#define OPTION_IMAGE_CENTER_STRING "image-center"
#define OPTION_IMAGE_SIZE 'z'
#define OPTION_IMAGE_SIZE_STRING "image-size"
#define OPTION_IMAGE_WINDOW 'w'
#define OPTION_IMAGE_WINDOW_STRING "image-window"
#define OPTION_INPUT_FILE 'I'
#define OPTION_INPUT_FILE_STRING "input-file"
#define OPTION_INTERPOLATION 'i'
#define OPTION_INTERPOLATION_STRING "interpolation"
#define OPTION_ISOCENTER 'o'
#define OPTION_ISOCENTER_STRING "isocenter"
#define OPTION_MULTISPECTRAL 'S'
#define OPTION_MULTISPECTRAL_STRING "multispectral"
#define OPTION_NUM_ANGLES 'a'
#define OPTION_NUM_ANGLES_STRING "num-angles"
#define OPTION_OUTPUT_FORMAT 't'
#define OPTION_OUTPUT_FORMAT_STRING "output-format"
#define OPTION_OUTPUT_PREFIX 'O'
#define OPTION_OUTPUT_PREFIX_STRING "output-prefix"
#define OPTION_RESOLUTION 'r'
#define OPTION_RESOLUTION_STRING "resolution"
#define OPTION_SCALE 's'
#define OPTION_SCALE_STRING "scale"
//#define OPTION_TRUE_PGM_STRING "true-pgm"
//#define OPTION_TRUE_PGM 't'

#define OUTPUT_FORMAT_PFM 0
#define OUTPUT_FORMAT_PGM 1
#define OUTPUT_FORMAT_RAW 2

#define IMGTYPE float
#define PREPROCESS_ATTENUATION 1

#define LINELEN 128
#define MAX_SHORT 32767
#define MIN_SHORT -32768

#define WRITE_BLOCK (1024*1024)

#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif
#ifndef M_TWOPI
#define M_TWOPI         (M_PI * 2.0)
#endif

typedef struct MGHDRR_Options_struct MGHDRR_Options;
struct MGHDRR_Options_struct {
	int image_resolution[2];         /* In pixels */
	float image_size[2];             /* In mm */
	int have_image_center;           /* Was image_center spec'd in options? */
	float image_center[2];           /* In pixels */
	int have_image_window;           /* Was image_window spec'd in options? */
	int image_window[4];             /* In pixels */
	float isocenter[3];              /* In mm */
	int num_angles;
	int have_angle_diff;             /* Was angle_diff spec'd in options? */
	float angle_diff;                /* In degrees */
	float sad;			     /* In mm */
	float sid;			     /* In mm */
	float scale;
	int exponential_mapping;
	int output_format;
	int multispectral;
	int interpolation;
	char* input_file;
	char* output_prefix;
};

enum Pixel_Type {
	PT_UNDEFINED,
	PT_UCHAR,
	PT_SHORT,
	PT_UINT32,
	PT_FLOAT,
	PT_VF_FLOAT_INTERLEAVED,
	PT_VF_FLOAT_PLANAR
};

enum point_location {
	POINTLOC_LEFT,
	POINTLOC_INSIDE,
	POINTLOC_RIGHT,
};
typedef enum point_location Point_location;

typedef struct volume Volume;
struct volume
{
	int dim[3];			// x, y, z Dims
	int npix;			// # of pixels in volume
	// = dim[0] * dim[1] * dim[2] 
	float offset[3];
	float pix_spacing[3];	// voxel spacing
	float direction_cosines[9];
	
	enum Pixel_Type pix_type;	// Voxel Data type
	int pix_size;		// (Unused?)
	void* img;			// Voxel Data
	
	float xmin;	// Minimum X Value in Volume
	float xmax; // Maximum X Value in Volume
	float ymin; // Minimum Y Value in Volume
	float ymax; // Maximum Y Value in Volume
	float zmin; // Minimum Z Value in Volume
	float zmax; // Maximum Z Value in Volume
};

struct volume_limit {
	double limits[2];   /* upper and lower limits of volume, including tolerances */
	int dir;		/* dir == 0 if limits go from low to high */
};
typedef struct volume_limit Volume_limit;

struct volume_limit_f {
	float limits[2];   /* upper and lower limits of volume, including tolerances */
	int dir;		/* dir == 0 if limits go from low to high */
};
typedef struct volume_limit_f Volume_limit_f;

// Prototyping

void accumulate_multispectral (float* bins, double pix_len, short pix_density);
float attenuation_lookup (float pix_density);
float attenuation_lookup_hu (float pix_density);
unsigned char bin_multispectral (short pix_density);
void dump_multispectral (FILE* msd_fp, float* bins);
void drr_render_volume_perspective (Volume* vol, double* cam, double* tgt, double* vup, double sid, double* ic, double* ps, int* ires, char* image_fn, char* multispectral_fn, MGHDRR_Options* options);
Point_location drr_test_boundary (Volume_limit* vol_limit, double x);
int drr_trace_init (int *ai_x, int *ai_y, int *ai_z, int *aixdir, int *aiydir, int *aizdir, double *ao_x, double *ao_y, double *ao_z, double *al_x, double *al_y, double *al_z, double *len, Volume* vol, Volume_limit vol_limits[3], double* p1, double* p2);
void drr_trace_init_loopvars_nointerp (int* ai, int* aidir, double* ao, double* al, double pt, double ry, double offset, double samp);
double drr_trace_ray_nointerp_2009 (Volume* vol, Volume_limit vol_limits[3], double* p1in, double* p2in, FILE* msd_fp);
double drr_trace_ray_trilin_approx (Volume* vol, double* p1in, double* p2in);
double drr_trace_ray_trilin_exact (Volume* vol, double* p1in, double* p2in);
void drr_write_projection_matrix (Volume* vol, double* cam, double* tgt, double* vup, double sid, double* ic, double* ps, int* ires, char* out_fn);
void drr_render_volumes (Volume* vol, MGHDRR_Options* options);
cl_ulong executionTime(cl_event &event);
void fwrite_block (void* buf, size_t size, size_t count, FILE* fp);
void init_multispectral (float* bins);
void parse_args (MGHDRR_Options* options, int argc, char* argv[]);
void preprocess_attenuation (Volume* vol);
void print_usage (void);
Volume* read_mha (char* filename);
void set_default_options (MGHDRR_Options* options);
void set_image_parms (MGHDRR_Options* options);
void set_isocenter (Volume* vol, MGHDRR_Options* options);
void volume_convert_to_float (Volume* ref);

float preprocess_attenuation_and_drr_render_volumes(Volume* vol, MGHDRR_Options* options);
void drr_render_volume_perspective_cl(Volume* vol, double* cam, double* tgt, double* vup, double sid, double* ic, double* ps, int* ires, char* image_fn, char* multispectral_fn,  MGHDRR_Options* options, int img_size, int multispectral_size, int valid_size, int float3_size, cl_mem *g_dev_vol, cl_mem *g_dev_img, cl_mem *g_dev_multispectral, cl_mem *g_dev_valid, cl_mem *c_vol_dim, cl_mem *c_img_dim, cl_mem *c_offset, cl_mem *c_pix_spacing, cl_mem *c_vol_limits, cl_mem *c_p1, cl_mem *c_ul_room, cl_mem *c_incr_r, cl_mem *c_incr_c, cl_mem *c_pixel_device, cl_kernel *drr_kernel, cl_ulong *drr_total, cl_ulong *img_total, cl_ulong *multispectral_total, cl_ulong *valid_total, size_t drr_global_work_size[MAX_GPU_COUNT][2], size_t drr_local_work_size[MAX_GPU_COUNT][2], int4 *pixels_per_device, int *offset, int *img_size_device, int *multispectral_size_device, int *valid_size_device);
#endif